/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.barista.api;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.Value;
import fr.x9c.cadmium.primitives.stdlib.Hash;
import fr.x9c.cadmium.support.values.ToValue;

/**
 * This class represents a descriptor for a Java element.
 *
 * @author <a href="mailto:barista@x9c.fr">Xavier Clerc</a>
 * @version 1.0
 * @since 1.0
 */
public abstract class Descriptor implements ToValue {

    /** Hash for 'Array' variant. */
    private static final Value HASH_ARRAY = Hash.hashVariant("Array");

    /** Hash for 'Class' variant. */
    private static final Value HASH_CLASS = Hash.hashVariant("Class");

    /** Map from tag values to flags. */
    private static final Map<Integer, Descriptor> MAP =
        new HashMap<Integer, Descriptor>();

    /**
     * Ensures that no descriptor could be defined outside this file.
     */
    private Descriptor() {
    } // end empty constructor

    /**
     * This class represents simple descriptor elements.
     */
    private static final class Simple extends Descriptor {

        /** Descriptor name. */
        private final String name;

        /** Print value. */
        private final String print;

        /**
         * Constructs a descriptor from its tag name.
         * @param n tag name - should not be <tt>null</tt>
         * @param p value to print - should not be <tt>null</tt>
         */
        private Simple(final String n, final String p) {
            assert n != null : "null n";
            assert p != null : "null p";
            this.name = n;
            this.print = p;
        } // end constructor(String, String)

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            return Hash.hashVariant(this.name);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return this.print;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.name.hashCode();
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof Simple) {
                final Simple that = (Simple) obj;
                return this.name.equals(that.name);
            } else {
                return false;
            } // end if/else
        } // end method 'toString()'

    } // end inner-class 'Simple'

    /** Descriptor for array type. */
    public static final class Array extends Descriptor {

        /** Elements type. */
        private final Descriptor type;

        /**
         * Constucts array descriptor from array element.
         * @param t element type - should be neither <tt>null</tt> nor <tt>VOID</tt>
         */
        public Array(final Descriptor t) {
            assert t != null : "null t";
            assert t != VOID : "t should not be VOID";
            this.type = t;
        } // end constructor(Descriptor)

        /**
         * Returns the type of array elements.
         * @return the type of array elements
         */
        public Descriptor getType() {
            return this.type;
        } // end method 'getType()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block res = Block.createBlock(0,
                                                HASH_ARRAY,
                                                this.type.toValue());
            return Value.createFromBlock(res);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return this.type.toString() + "[]";
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.type.hashCode();
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof Array) {
                final Array that = (Array) obj;
                return this.type.equals(that.type);
            } else {
                return false;
            } // end if/else
        } // end method 'toString()'

    } // end inner-class 'Array'

    /** Descriptor for boolean type. */
    public static final Descriptor BOOLEAN = new Simple("Boolean", "boolean");

    /** Descriptor for byte type. */
    public static final Descriptor BYTE = new Simple("Byte", "byte");

    /** Descriptor for char type. */
    public static final Descriptor CHAR = new Simple("Char", "char");

    /** Descriptor for class type. */
    public static final class Class extends Descriptor {

        /** Class name. */
        private final String name;

        /**
         * Constucts class name from string.
         * @param n class name - should not be <tt>null</tt>
         */
        public Class(final String n) {
            assert n != null : "null n";
            this.name = n;
        } // end constructor(String)

        /**
         * Returns the class name.
         * @return the class name
         */
        public String getName() {
            return this.name;
        } // end method 'getName()'

        /**
         * {@inheritDoc}
         */
        public Value toValue() {
            final Block res = Block.createBlock(0,
                                                HASH_CLASS,
                                                API.createClassName(this.name));
            return Value.createFromBlock(res);
        } // end method 'toValue()'

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return this.name;
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public int hashCode() {
            return this.name.hashCode();
        } // end method 'toString()'

        /**
         * {@inheritDoc}
         */
        @Override
        public boolean equals(final Object obj) {
            if (obj instanceof Class) {
                final Class that = (Class) obj;
                return this.name.equals(that.name);
            } else {
                return false;
            } // end if/else
        } // end method 'toString()'

    } // end inner-class 'Class'

    /** Descriptor for double type. */
    public static final Descriptor DOUBLE = new Simple("Double", "double");

    /** Descriptor for float type. */
    public static final Descriptor FLOAT = new Simple("Float", "float");

    /** Descriptor for int type. */
    public static final Descriptor INT = new Simple("Int", "int");

    /** Descriptor for long type. */
    public static final Descriptor LONG = new Simple("Long", "int");

    /** Descriptor for short type. */
    public static final Descriptor SHORT = new Simple("Short", "short");

    /** Descriptor for void type. */
    public static final Descriptor VOID = new Simple("Void", "void");

    /**
     * Tests whether the descriptor can be a field descriptor.
     * @return <tt>true</tt> if the descriptor can be a field descriptor,
     *         <tt>false</tt> otherwise
     */
    public final boolean isFieldDescriptor() {
        return this != VOID;
    } // end method 'isFieldDescriptor()'

    /**
     * Tests whether a list contains only field descriptors.
     * @param l list to test - should not be <tt>null</tt>
     * @return <tt>true</tt> if the list contains only field descriptors,
     *         <tt>false</tt> otherwise
     */
    public static boolean areFieldDescriptors(final List<? extends Descriptor> l) {
        assert l != null : "null l";
        for (Descriptor d : l) {
            if (!d.isFieldDescriptor()) {
                return false;
            } // end if
        } // end for
        return true;
    } // end method 'areFieldDescriptors(List<? extends Descriptor>)'

    /**
     * Constructs an instance from a value.
     * @param v value to construct instance from - should not be <tt>null</tt>
     * @return the corresponding instance
     */
    static Descriptor fromValue(final Value v) {
        assert v != null : "null v";
        if (v.isLong()) {
            return MAP.get(v.asLong());
        } else {
            final Block b = v.asBlock();
            final int tag = b.get(0).asLong();
            if (tag == HASH_ARRAY.asLong()) {
                return new Array(fromValue(b.get(1)));
            } else if (tag == HASH_CLASS.asLong()) {
                return new Class(API.stringOfClassName(b.get(1)));
            } else {
                assert false : "invalid descriptor";
                return null;
            } // end if/elsif/else
        } // end if/else
    } // end method 'fromValue(Value)'

    /**
     * Converts a descriptor list into a string.
     * @param l list to convert - should not be <tt>null</tt>
     * @return the concatenation of all list elements
     */
    public static String listToString(final List<? extends Descriptor> l) {
        assert l != null : "null l";
        final StringBuilder sb = new StringBuilder();
        boolean first = true;
        for (Descriptor d : l) {
            if (first) {
                first = false;
            } else {
                sb.append(" ");
            } // end if/else
            sb.append(d.toString());
        } // end for
        return sb.toString();
    } // end method 'listToString(List<? extends Descriptor>)'

    /** Map initilization. */
    static {
        MAP.put(BOOLEAN.toValue().asLong(), BOOLEAN);
        MAP.put(BYTE.toValue().asLong(), BYTE);
        MAP.put(CHAR.toValue().asLong(), CHAR);
        MAP.put(DOUBLE.toValue().asLong(), DOUBLE);
        MAP.put(FLOAT.toValue().asLong(), FLOAT);
        MAP.put(INT.toValue().asLong(), INT);
        MAP.put(LONG.toValue().asLong(), LONG);
        MAP.put(SHORT.toValue().asLong(), SHORT);
        MAP.put(VOID.toValue().asLong(), VOID);
    } // end static block

} // end class 'Descriptor'
