(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

let register_callbacks () =
  let string_of_exception = function
    | AccessFlag.Exception e -> AccessFlag.string_of_error e
    | Annotation.Exception e -> Annotation.string_of_error e
    | Assembler.Exception (l, e) -> Printf.sprintf "%d: %s" l (Assembler.string_of_error e)
    | Attribute.Exception e -> Attribute.string_of_error e
    | ByteCode.Exception e -> ByteCode.string_of_error e
    | ClassDefinition.Exception e -> ClassDefinition.string_of_error e
    | ClassFile.Exception e -> ClassFile.string_of_error e
    | ClassLoader.Exception e -> ClassLoader.string_of_error e
    | ClassPath.Exception e -> ClassPath.string_of_error e
    | ConstantPool.Exception e -> ConstantPool.string_of_error e
    | ControlFlow.Exception e -> ControlFlow.string_of_error e
    | Descriptor.Exception e -> Descriptor.string_of_error e
    | Field.Exception e -> Field.string_of_error e
    | InputStream.Exception e -> InputStream.string_of_error e
    | Instruction.Exception e -> Instruction.string_of_error e
    | Lexer.Exception e -> Lexer.string_of_error e
    | Method.Exception e -> Method.string_of_error e
    | Name.Exception e -> Name.string_of_error e
    | OutputStream.Exception e -> OutputStream.string_of_error e
    | Signature.Exception e -> Signature.string_of_error e
    | Source.Exception e -> Source.string_of_error e
    | StackState.Exception e -> StackState.string_of_error e
    | Utils.Unicode_exception e -> Utils.string_of_unicode_error e
    | Utils.Lexer_exception e -> Utils.string_of_lexer_error e
    | Version.Exception e -> Version.string_of_error e
    | e -> Printexc.to_string e in
  let reg = Callback.register in
  let reg1 n f =
    Callback.register
      n 
      (fun x -> try f x with e -> failwith (string_of_exception e)) in
  reg1 "fr.x9c.barista.api.Name.make_class" Name.make_for_class_from_external;
  reg1 "fr.x9c.barista.api.Name.make_field" Name.make_for_field;
  reg1 "fr.x9c.barista.api.Name.make_method" Name.make_for_method;
  reg1 "fr.x9c.barista.api.Name.utf8_of_class" Name.external_utf8_for_class;
  reg1 "fr.x9c.barista.api.Name.utf8_of_field" Name.utf8_for_field;
  reg1 "fr.x9c.barista.api.Name.utf8_of_method" Name.utf8_for_method;
  reg1 "fr.x9c.barista.api.Signature.make_field" Signature.field_type_signature_of_utf8;
  reg1 "fr.x9c.barista.api.Signature.utf8_of_field_signature" Signature.utf8_of_field_type_signature;
  reg1 "fr.x9c.barista.api.Signature.make_method" Signature.method_signature_of_utf8;
  reg1 "fr.x9c.barista.api.Signature.utf8_of_method_signature" Signature.utf8_of_method_signature;
  reg1 "fr.x9c.barista.api.Signature.make_class" Signature.class_signature_of_utf8;
  reg1 "fr.x9c.barista.api.Signature.utf8_of_class_signature" Signature.utf8_of_class_signature;
  reg "fr.x9c.barista.api.Printer.print"
    (fun n c o ->
      try
        Printer.print_to_channel o (ClassPath.make_of_string c) n
      with e -> failwith (string_of_exception e));
  reg "fr.x9c.barista.api.Disassembler.disassemble"
    (fun n c o ->
      try
        Disassembler.disassemble_to_channel o (ClassPath.make_of_string c) n
      with e -> failwith (string_of_exception e));
  reg "fr.x9c.barista.api.Assembler.assemble"
    (fun v s c i o ->
      try
        Name.external_utf8_for_class
          (Assembler.assemble
             ~version:v
             ~compute_stacks:s
             ~class_loader:(ClassLoader.make (ClassPath.make_of_string c))
             i
             o)
      with e -> failwith (string_of_exception e));
  reg "fr.x9c.barista.api.ClassDefinition.encode"
    (fun v cd ->
      try
        let buff = Buffer.create 256 in
        let os = OutputStream.make_of_buffer buff in
        ClassFile.write (ClassDefinition.encode ~version:v cd) os;
        OutputStream.close os;
        Buffer.contents buff
      with e -> failwith (string_of_exception e));
  reg "fr.x9c.barista.api.ClassDefinition.decode"
    (fun ic ->
      try
        let is = InputStream.make_of_channel ic in
        ClassDefinition.decode (ClassFile.read is)
      with e -> failwith (string_of_exception e))
