(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(** This module defines methods in both low- and high-level forms.
   It also provides conversion functions between levels as well as i/o
   functions for low-level. *)


(** {6 Low-level form} *)

type info = {
    access_flags : Utils.u2;
    name_index : Utils.u2;
    descriptor_index : Utils.u2;
    attributes_count : Utils.u2;
    attributes : Attribute.info array;
  }
(** Represents a method as defined in the class file format specification. *)


(** {6 Exception} *)

type error =
  | Invalid_name
  | Invalid_name_value
  | Invalid_descriptor_value

exception Exception of error
(** Exception to be raised when a function of this module fails.  *)

val string_of_error : error -> string
(** Converts the passed error into a string. *)


(** {6 I/O functions} *)

val read_info : InputStream.t -> info
(** [read_info st] reads a method from [st].
   Raises [InputStream.Exception] if an i/o error occurs. *)

val write_info : OutputStream.t -> info -> unit
(** [write_info st m] writes method [m] onto [st].
   Raises [OutputStream.Exception] if an i/o error occurs. *)


(** {6 High-level form} *)

type t =
| Regular of (AccessFlag.for_method list) * Name.for_method * Descriptor.for_method * (Attribute.for_method list) (** Regular method (flags, name, descriptor, and attributes). *)
  | Constructor of (AccessFlag.for_constructor list) * (Descriptor.for_parameter list) * (Attribute.for_method list) (** Instance initializer (flags, descriptor, and attributes). *)
  | Initializer of bool * (Attribute.for_method list) (** Class initializer (attributes, the boolean indicating whether the initializer is {i strictfp}). *)
(** Represents the different kinds of methods. *)


(** {6 Conversion functions} *)

val decode : bool -> ConstantPool.t -> info -> t
(** Converts from a low-level into a high-level form according to passed pool.
   The passed boolean indicates whether the enclosing element is an interface.
   Raises [Exception] if an error occurs during conversion. *)

val encode : ConstantPool.extendable -> t -> info
(** Converts from a high-level into a low-level form, using passed extendable
   pool.
   Raises [Exception] if an error occurs during conversion. *)
