(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

open Utils
open Consts


(* Exception *)

type error =
  | Invalid_array_dimensions
  | Invalid_type
  | Unknown_flag

exception Exception of error

let fail e = raise (Exception e)

let string_of_error = function
  | Invalid_array_dimensions -> "invalid number of dimensions for array"
  | Invalid_type -> "invalid type"
  | Unknown_flag -> "unknown flag"

let () =
  Printexc.register_printer
    (function
      | Exception e -> Some (string_of_error e)
      | _ -> None)


(* Orders over some Java elements *)

let flag_list_order fl1 fl2 =
  let s1 = List.mem `Static fl1 in
  let s2 = List.mem `Static fl2 in
  let pub1 = List.mem `Public fl1 in
  let pub2 = List.mem `Public fl2 in
  let pro1 = List.mem `Protected fl1 in
  let pro2 = List.mem `Protected fl2 in
  let pri1 = List.mem `Private fl1 in
  let pri2 = List.mem `Private fl2 in
  compare (pub2, pro2, pri2, s2) (pub1, pro1, pri1, s1)

let flag_order f1 f2 =
  let flag_rank = function
  | `Public
  | `Private
  | `Protected -> 1
  | `Static -> 2
  | `Final -> 3
  | `Synchronized -> 4
  | `Volatile -> 4
  | `Transient -> 4
  | `Native -> 4
  | `Abstract -> 2
  | `Strict -> 4
  | `Super
  | `Bridge
  | `Varargs
  | `Interface
  | `Synthetic
  | `Annotation
  | `Enum -> 5 in
  let cmp = compare (flag_rank f1) (flag_rank f2) in
  if cmp = 0 then compare f1 f2 else cmp

let field_order f1 f2 =
  let fl1, n1, _, _ = f1 in
  let fl2, n2, _, _ = f2 in
  let cmp = flag_list_order fl1 fl2 in
  if cmp <> 0 then
    cmp
  else
    let cmp2 = utf8_compare (Name.utf8_for_field n1) (Name.utf8_for_field n2) in
    if cmp2 <> 0 then
      cmp2
    else
      compare f1 f2

let meth_desc_order md1 md2 =
  let p1, r1 = md1 in
  let p2, r2 = md2 in
  let cmp = compare (List.length p1) (List.length p2) in
  if cmp <> 0 then
    cmp
  else
    let cmp' = compare (Descriptor.utf8_of_method md1) (Descriptor.utf8_of_method md2) in
    if cmp' <> 0 then
      cmp'
    else
      compare (Descriptor.utf8_of_java_type r1) (Descriptor.utf8_of_java_type r2)

let method_order m1 m2 =
  let method_rank = function
    | Method.Regular _ -> 2
    | Method.Constructor _ -> 1
    | Method.Initializer _ -> 0 in
  let r1 = method_rank m1 in
  let r2 = method_rank m2 in
  let cmp = compare r1 r2 in
  if cmp <> 0 then
    cmp
  else
    match m1, m2 with
    | Method.Regular (fl1, n1, d1, _), Method.Regular (fl2, n2, d2, _) ->
        let cmp' = flag_list_order fl1 fl2 in
        if cmp' <> 0 then
          cmp'
        else
          let cmp'' = utf8_compare (Name.utf8_for_method n1) (Name.utf8_for_method n2) in
          if cmp'' <> 0 then
            cmp''
          else
            meth_desc_order d1 d2
    | Method.Constructor (fl1, d1, _), Method.Constructor (fl2, d2, _) ->
        let cmp = compare (List.mem `Public fl1) (List.mem `Public fl2) in
        if cmp <> 0 then
          cmp
        else
          meth_desc_order (d1, `Void) (d2, `Void)
    | _ -> compare m1 m2

let attribute_order a1 a2 =
  let attribute_rank = function
    | `ConstantValue _ -> 3
    | `Code _ -> 8
    | `Exceptions _ -> 3
    | `InnerClasses _ -> 2
    | `EnclosingMethod _ -> 2
    | `Synthetic -> 0
    | `Signature _ -> 1
    | `SourceFile _ -> 1
    | `SourceDebugExtension _ -> 6
    | `LineNumberTable _ -> -1
    | `LocalVariableTable _ -> -1
    | `LocalVariableTypeTable _ -> -1
    | `Deprecated -> 0
    | `RuntimeVisibleAnnotations _ -> 5
    | `RuntimeInvisibleAnnotations _ -> 5
    | `RuntimeVisibleParameterAnnotations _ -> 5
    | `RuntimeInvisibleParameterAnnotations _ -> 5
    | `AnnotationDefault _ -> 4
    | `StackMapTable _ -> 9
    | `Unknown _ -> 7 in
  let cmp = compare (attribute_rank a1) (attribute_rank a2) in
  if cmp = 0 then compare a1 a2 else cmp


(* Some conversion functions *)

let rec utf8_of_java_type = function
  | `Boolean -> utf8_of_string "boolean"
  | `Byte -> utf8_of_string "byte"
  | `Char -> utf8_of_string "char"
  | `Double -> utf8_of_string "double"
  | `Float -> utf8_of_string "float"
  | `Int -> utf8_of_string "int"
  | `Long -> utf8_of_string "long"
  | `Short -> utf8_of_string "short"
  | `Void -> utf8_of_string "void"
  | `Class n -> Name.printable_utf8_for_class n
  | `Array jt -> (utf8_of_java_type (jt :> Descriptor.java_type)) ++ (utf8_of_string "[]")

let java_type_of_utf8 s =
  let rec make_array n x =
    if n = 0 then
      x
    else
      `Array (make_array (pred n) x) in
  let l = utf8_length s in
  let i = ref 0 in
  while !i < l && ((uchar_is_letter_or_digit (utf8_get s !i))
                 || (uchar_eq dot (utf8_get s !i))
                 || (uchar_eq opening_square_bracket (utf8_get s !i))
                 || (uchar_eq closing_square_bracket (utf8_get s !i))) do
    incr i
  done;
  if !i = l && uchar_is_letter (utf8_get s 0) then
    let j = ref (pred l) in
    let dims = ref 0 in
    while (!j - 1 >= 0)
        && (uchar_eq closing_square_bracket (utf8_get s !j))
        && (uchar_eq opening_square_bracket (utf8_get s (!j - 1))) do
      incr dims;
      decr j;
      decr j
    done;
    if !dims > 255 then fail Invalid_array_dimensions;
    let prefix = utf8_substring s 0 !j in
    let base = match (try string_of_utf8 prefix with _ -> "") with
    | "boolean" -> `Boolean
    | "byte" -> `Byte
    | "char" -> `Char
    | "double" -> `Double
    | "float" -> `Float
    | "int" -> `Int
    | "long" -> `Long
    | "short" -> `Short
    | "void" -> `Void
    | _ -> `Class (Name.make_for_class_from_external prefix) in
    if !dims = 0 then
      base
    else
      let array = make_array !dims (Descriptor.filter_void Descriptor.Invalid_array_element_type base) in
      (array :> Descriptor.java_type)
  else
    fail Invalid_type

let java_type_of_utf8_no_void s =
  let res = java_type_of_utf8 s in
  Descriptor.filter_void Descriptor.Void_not_allowed res

let utf8_of_flag f =
  let uof = function
    | `Public -> "public"
    | `Private -> "private"
    | `Protected -> "protected"
    | `Static -> "static"
    | `Final -> "final"
    | `Super -> "super"
    | `Synchronized -> "synchonized"
    | `Bridge -> "bridge"
    | `Volatile -> "volatile"
    | `Transient -> "transient"
    | `Varargs -> "varargs"
    | `Native -> "native"
    | `Interface -> "interface"
    | `Abstract -> "abstract"
    | `Strict -> "strictfp"
    | `Synthetic -> "synthetic"
    | `Annotation -> "annotation"
    | `Enum -> "enum" in
  utf8_of_string (uof f)

let utf8_of_flags l =
  (utf8_concat_sep
    (utf8_of_string " ")
    (List.map utf8_of_flag (List.sort flag_order l)))
    ++ (if l <> [] then utf8_of_string " " else empty_utf8)

let flag_of_utf8 s =
  let fou = function
    | "public" -> `Public
    | "private" -> `Private
    | "protected" -> `Protected
    | "static" -> `Static
    | "final" -> `Final
    | "super" -> `Super
    | "synchonized" -> `Synchronized
    | "bridge" -> `Bridge
    | "volatile" -> `Volatile
    | "transient" -> `Transient
    | "varargs" -> `Varargs
    | "native" -> `Native
    | "interface" -> `Interface
    | "abstract" -> `Abstract
    | "strictfp" -> `Strict
    | "synthetic" -> `Synthetic
    | "annotation" -> `Annotation
    | "enum" -> `Enum
    | _ -> fail Unknown_flag in
  fou (string_of_utf8 s)


(* Miscellaneous *)

let output chan buf =
  let out = utf8_line_writer_make chan in
  utf8_line_writer_put out (utf8_buf_contents buf);
  utf8_line_writer_flush out
