(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

(** This module provides some utility functions used throughout the Barista
   project. In particular, it defines type and functions for UTF8 strings.
   UTF8-related types are abstract and rely on Camomile
   (available at http://camomile.sourceforge.net). *)


(** {6 Integer types} *)

(** The following aliases are defined to keep type expressions as close as
   possible from Java specification. *)

type u1 = int
(** The type of unsigned 1-byte integers. *)

type s1 = int
(** The type of signed 1-byte integers. *)

type u2 = int
(** The type of unsigned 2-byte integers. *)

type s2 = int
(** The type of signed 2-byte integers. *)

type u4 = int64
(** The type of unsigned 4-byte integers. *)

type s4 = int32
(** The type of signed 4-byte integers. *)


(** {6 Exceptions} *)

type unicode_error =
  | Unrepresentable_character
  | Invalid_character_code
  | Unable_to_convert_to_modified_utf8
  | Unable_to_convert_from_modified_utf8
  | Unable_to_convert_to_utf8
  | Unable_to_convert_from_utf8
  | Invalid_index
  | Index_out_of_bounds
  | Invalid_escaped_string

exception Unicode_exception of unicode_error
(** Exception to be thrown when conversion from/to Unicode fails. *)

val string_of_unicode_error : unicode_error -> string
(** Converts the passed error into a string. *)

type lexer_error =
  | End_of_lexer
  | Invalid_consume of char * char

exception Lexer_exception of lexer_error
(** Exception to be thrown when a try is made to consume an invalid character
   from a lexer or when end of lexer is unexpectedly reached. *)

val string_of_lexer_error : lexer_error -> string
(** Converts the passed error into a string. *)


(** {6 Unicode characters} *)

type uchar
(** The type of Unicode characters. *)

val uchar_of_char : char -> uchar
(** Conversion from standard character into Unicode character. *)

val char_of_uchar : uchar -> char
(** Conversion from Unicode character into standard character.
   Raises [Unicode_exception] if conversion fails. *)

val uchar_of_code : int -> uchar
(** Conversion from Unicode code value into Unicode character.
   Raises [Unicode_exception] if conversion fails. *)

val code_of_uchar : uchar -> int
(** Conversion from Unicode character into Unicode code value.
   Raises [Unicode_exception] if conversion fails. *)

val uchar_eq : uchar -> uchar -> bool
(** Equality over Unicode characters. *)

val uchar_is_letter : uchar -> bool
(** Predicate testing whether the passed char is a letter. *)

val uchar_is_digit : uchar -> bool
(** Predicate testing whether the passed char is a digit. *)

val uchar_is_letter_or_digit : uchar -> bool
(** Predicate testing whether the passed char is a letter, a digit, or an underscore. *)


(** {6 UTF8 strings} *)

type utf8
(** The type of UTF8 strings. *)

type modified
(** The type of UTF8 strings in {i modified} format, as they appear in Java
   class files.
    cf. {i http://java.sun.com/javase/6/docs/api/java/io/DataInput.html#modified-utf-8}*)

val make_utf8 : uchar list -> utf8
(** Constructs an UTF8 string from a list of Unicode characters. *)

val modified_of_bytes : string -> modified
(** Conversion from bytes into modified UTF8 string. *)

val bytes_of_modified : modified -> string
(** Conversion from modified UTF8 string into bytes. *)

val modified_of_utf8 : utf8 -> modified
(** Conversion from UTF8 string into modified UTF8 string.
   Raises [Unicode_exception] if conversion fails. *)

val utf8_of_modified : modified -> utf8
(** Conversion from modified UTF8 string into UTF8 string.
   Raises [Unicode_exception] if conversion fails. *)

val string_of_utf8 : utf8 -> string
(** Conversion from UTF8 string into string.
   Raises [Unicode_exception] if conversion fails. *)

val utf8_of_string : string -> utf8
(** Conversion from string into UTF8 string.
   Raises [Unicode_exception] if conversion fails. *)

val utf8_length : utf8 -> int
(** Returns the length of the passed string (in characters, not bytes). *)

val utf8_get : utf8 -> int -> uchar
(** [utf8_get s i] returns the character of [s] at index [i].
   Raises [Unicode_exception] if index is not valid.*)

val utf8_eq : utf8 -> utf8 -> bool
(** Equality over UTF8 strings. *)

val utf8_compare : utf8 -> utf8 -> int
(** Comparison over UTF8 strings. *)

val utf8_index_from : utf8 -> int -> uchar -> int
(** [utf8_index_from s i c] returns the lowest index above or equal to [i]
   where string [s] contains a character that equals [c].
   Raises [Not_found] if such an index does not exist. *)

val utf8_rindex_from : utf8 -> int -> uchar -> int
(** [utf8_rindex_from s i c] returns the highest index below or equal to [i]
   where string [s] contains a character that equals [c].
   Raises [Not_found] if such an index does not exist. *)

val utf8_substring : utf8 -> int -> int -> utf8
(** [utf8_substring s start end] returns a string whose characters are those
   of [s] from index [first] to index [last] (both inclusive).
   Returns an empty string iff [last < first].
   Raises [Unicode_exception] if [first] or [last] is not a valid index. *)

val (++) : utf8 -> utf8 -> utf8
(** Concatenation of UTF8 strings. *)

val utf8_concat : utf8 list -> utf8
(** [utf8_concat l] returns the concatenation of all strings in [l]. *)

val utf8_concat_sep : utf8 -> utf8 list -> utf8
(** [utf8_concat sep l] returns the concatenation of all strings in [l],
   separator [sep] being inserted between two strings. *)

val utf8_replace : uchar -> uchar -> utf8 -> utf8
(** [utf8_replace c1 c2 s] returns a copy of [s] where every character equal to
   [c1] has been replaced by [c2]. *)

val utf8_contains : uchar -> utf8 -> bool
(** [utf8_contains c s] returns [true] iff string [s] contains a character
   equal to [c]. *)

val utf8_split : uchar -> utf8 -> utf8 list
(** [utf8_split c s] returns the string list obtained by splitting [s] using
   delimiter [c]. *)

val utf8_escape : utf8 -> utf8
(** [utf8_escape s] returns the constant string corresponding to the passed
   string. A leading and a trailing quote are added and every quote inside
   the passed string is escaped by a back slash. *)

val utf8_unescape : utf8 -> utf8
(** [utf8_unescape s] returns the string corresponding to the passed constant
   string. Both leading and trailing quotes are removed and escaped sequences
   are converted. *)

val utf8_escape_char : uchar -> utf8
(** [utf8_escape_char c] returns the constant string corresponding to the
   passed character. A leading and a trailing simple quote are added. *)


(** {6 UTF8 hash tables} *)

module UTF8Hashtbl : Hashtbl.S with type key = utf8


(** {6 UTF8 buffers} *)

type utf8_buf
(** The type of UTF8 buffers. *)

val utf8_buf_make : unit -> utf8_buf
(** Constructs an empty buffer. *)

val utf8_buf_add_char : utf8_buf -> uchar -> unit
(** [utf8_buf_add_char b c] appends char [c] to buffer [b]. *)

val utf8_buf_add_string : utf8_buf -> utf8 -> unit
(** [utf8_buf_add_char b s] appends string [s] to buffer [b]. *)

val utf8_buf_add_endline : utf8_buf -> utf8 -> unit
(** [utf8_buf_add_char b s] appends string [s] to buffer [b]
   and then also appends the end-of-line character. *)

val utf8_buf_contents : utf8_buf -> utf8
(** [utf8_buf_contents b] returns the contents of buffer [b] as a string. *)


(** {6 UTF8 line-oriented readers} *)

type utf8_line_reader
(** The type of UTF8 line-oriented readers. *)

val utf8_line_reader_make : in_channel -> utf8_line_reader
(** Constructs a line reader from an input channel. *)

val utf8_line_reader_get : utf8_line_reader -> utf8
(** [utf8_line_reader_get r] returns the next line read from [r].
   Raises [End_of_file] if end of file is encountered. *)

val utf8_line_reader_close : utf8_line_reader -> unit
(** Closes the passed reader. *)


(** {6 UTF8 line-oriented writers} *)

type utf8_line_writer
(** The type of UTF8 line-oriented writers. *)

val utf8_line_writer_make : out_channel -> utf8_line_writer
(** Constructs a line writer from an output channel. *)

val utf8_line_writer_put : utf8_line_writer -> utf8 -> unit
(** [utf8_line_writer_put w s] outputs string [s] onto writer [w]. *)

val utf8_line_writer_flush : utf8_line_writer -> unit
(** Flushes the passed writer. *)

val utf8_line_writer_close : utf8_line_writer -> unit
(** Closes the passed writer. *)


(** {6 State objects for lexers over UTF8 strings} *)

class lexer_state : utf8 ->
  object
    method consume : unit
        (** Consumes (that is skips) the next character.
           Raises [Lexer_exception] if end of string is encountered. *)
    method consume_char : uchar
        (** Consumes the next character and returns it.
           Raises [Lexer_exception] if end of string is encountered. *)
    method consume_only : uchar -> unit
        (** Consumes (that is skips) the next character only if it is equal
           to the passed character.
           Raises [Lexer_exception] if end of string is encountered or if
           passed character is not equal to the next one of the lexer. *)
    method consume_until : uchar -> utf8
        (** Consumes characters until a character equal to the passed one is
           read from the lexer and returns the string of consumed characters
           (the passed character is not consumed).
           Raises [Lexer_exception] if end of string is encountered. *)
    method consume_until_list : uchar list -> utf8
        (** Consumes characters until a character equal to one of the
           passed list is read from the lexer and returns the string of
           consumed characters (the passed characters are not consumed).
           Raises [Lexer_exception] if end of string is encountered. *)
    method is_available : bool
        (** Returns [true] iff there is at least one character left on the
           lexer. *)
    method look_ahead : uchar -> bool
        (** Returns [true] iff the next character of the lexer is equal to the
           passed one.
           Raises [Lexer_exception] if end of string is encountered. *)
    method look_ahead_list : uchar list -> bool
        (** Returns [true] iff the next character of the lexer is equal to one
           of the passed list.
           Raises [Lexer_exception] if end of string is encountered. *)
  end
(** This class encapsulates the state of a lexer over an UTF8 string that is
   passed at instance creation. *)
